package ru.dbotthepony.mc.otm.data

import com.google.gson.JsonObject
import com.mojang.datafixers.util.Pair
import com.mojang.serialization.Codec
import com.mojang.serialization.DataResult
import com.mojang.serialization.DynamicOps
import com.mojang.serialization.JsonOps
import com.mojang.serialization.MapCodec
import com.mojang.serialization.MapLike
import com.mojang.serialization.RecordBuilder
import net.minecraft.advancements.critereon.ContextAwarePredicate
import net.minecraft.advancements.critereon.DeserializationContext
import net.minecraft.advancements.critereon.EntityPredicate
import ru.dbotthepony.mc.otm.core.fromJsonStrict
import ru.dbotthepony.mc.otm.core.set
import ru.dbotthepony.mc.otm.core.toJsonStrict
import java.util.Optional
import java.util.stream.Stream

class Codec2TriggerSerializer<S : Any>(supplier: (Codec2TriggerSerializer<S>.Context) -> Codec<S>) : Codec<S> {
	private val codec = supplier.invoke(Context())

	override fun <T : Any> encode(input: S, ops: DynamicOps<T>, prefix: T): DataResult<T> {
		return codec.encode(input, ops, prefix)
	}

	override fun <T : Any> decode(ops: DynamicOps<T>, input: T): DataResult<Pair<S, T>> {
		return codec.decode(ops, input)
	}

	fun fromJson(element: JsonObject, player: Optional<ContextAwarePredicate>, context: DeserializationContext): S {
		try {
			Companion.context.get().addLast(ThreadContext(player, context))
			return codec.fromJsonStrict(element)
		} finally {
			Companion.context.get().removeLast()
		}
	}

	fun toJson(value: S): JsonObject {
		return codec.toJsonStrict(value) as JsonObject
	}

	inner class Context {
		val playerPredicate: MapCodec<Optional<ContextAwarePredicate>>
			get() = CAPPredicate

		val awareContextCodec: Codec<ContextAwarePredicate>
			get() = ActualCAPPredicate
	}

	private data class ThreadContext(val player: Optional<ContextAwarePredicate>, val context: DeserializationContext)

	private object ActualCAPPredicate : Codec<ContextAwarePredicate> {
		override fun <T : Any> encode(input: ContextAwarePredicate, ops: DynamicOps<T>, prefix: T): DataResult<T> {
			return try {
				DataResult.success(JsonOps.INSTANCE.convertTo(ops, input.toJson()))
			} catch (err: Exception) {
				DataResult.error { "Failed to serialize ContextAwarePredicate: " + err.message }
			}
		}

		override fun <T : Any> decode(ops: DynamicOps<T>, input: T): DataResult<Pair<ContextAwarePredicate, T>> {
			val context = context.get().lastOrNull() ?: return DataResult.error { "Not current deserializing trigger instance" }

			return try {
				DataResult.success(Pair(EntityPredicate.fromJson(JsonObject().also { it["a"] = ops.convertTo(JsonOps.INSTANCE, input) }, "a", context.context).get(), ops.empty()))
			} catch (err: Exception) {
				DataResult.error { "Failed to deserialize ContextAwarePredicate: " + err.message }
			}
		}
	}

	private object CAPPredicate : MapCodec<Optional<ContextAwarePredicate>>() {
		override fun <T : Any> keys(ops: DynamicOps<T>): Stream<T> {
			return Stream.of(ops.createString("player"))
		}

		override fun <T : Any> decode(ops: DynamicOps<T>, input: MapLike<T>): DataResult<Optional<ContextAwarePredicate>> {
			return DataResult.success(context.get().lastOrNull()?.player ?: return DataResult.error { "Not currently deserializing trigger instance" })
		}

		override fun <T : Any> encode(input: Optional<ContextAwarePredicate>, ops: DynamicOps<T>, prefix: RecordBuilder<T>): RecordBuilder<T> {
			if (input.isPresent) {
				return prefix.add("player", JsonOps.INSTANCE.convertTo(ops, input.get().toJson()))
			} else {
				return prefix
			}
		}
	}

	companion object {
		private val context = object : ThreadLocal<ArrayDeque<ThreadContext>>() {
			override fun initialValue(): ArrayDeque<ThreadContext> {
				return ArrayDeque()
			}
		}
	}
}
